"""Kepware device module"""
from keppy.tag_group import TagGroup
from keppy.tag import Tag

class Device(object):
    """Represents a Kepware device"""
    def __init__(self, device_dict, is_sixteen_bit, ignore_list):
        self._ignore_list = ignore_list
        self._device_dict = device_dict
        self._tag_groups = self.parse_tag_groups()
        self._tags = self.parse_tags()
        self._is_sixteen_bit = is_sixteen_bit
        self.set_driver_simulated()

    def set_driver_simulated(self):
        """Sets the device driver type to simulated"""
        self._device_dict["servermain.MULTIPLE_TYPES_DEVICE_DRIVER"] = "Simulator"
        if self._is_sixteen_bit:
            self._device_dict["servermain.DEVICE_MODEL"] = 0
        else:
            self._device_dict["servermain.DEVICE_MODEL"] = 1
        self._device_dict["servermain.DEVICE_ID_OCTAL"] = 1

    def parse_tag_groups(self):
        """Gets an array of TagGroup objects in the Kepware device"""
        tag_groups = []
        if 'tag_groups' not in self._device_dict:
            return tag_groups
        to_remove = []
        for tag_group in self._device_dict['tag_groups']:
            if tag_group['common.ALLTYPES_NAME'] in self._ignore_list:
                to_remove.append(tag_group)
                continue
            tag_groups.append(TagGroup(tag_group))
        for removable in to_remove:
            self._device_dict['tag_groups'].remove(removable)

        return tag_groups

    def parse_tags(self):
        tags = []
        if 'tags' not in self._device_dict:
            return tags
        for tag in self._device_dict['tags']:
            tags.append(Tag(tag))
        return tags

    @property
    def tag_groups(self):
        """Gets the tag groups of the device"""
        return self._tag_groups

    @property
    def tags(self):
        return self._tags

    @property
    def is_sixteen_bit(self):
        """Returns true if is sixteen bit device"""
        return self._is_sixteen_bit

    @property
    def name(self):
        """Gets the name of the device"""
        return self._device_dict["common.ALLTYPES_NAME"]

    def as_dict(self):
        """Returns dictionary representation of the device"""
        return self._device_dict

    def update(self):
        """Updates the dictionary of the device"""
        if "tag_groups" not in self._device_dict:
            return
        for group in self.tag_groups:
            group.update()
        for i in range(len(self._device_dict["tag_groups"])):
            tag_group_dict = self._device_dict["tag_groups"][i]
            for group in self.tag_groups:
                if group.name == tag_group_dict["common.ALLTYPES_NAME"]:
                    self._device_dict["tag_groups"][i] = group.as_dict()



"""Represents a string register (SXXX)"""
from keppy.register import Register

class StringRegister(Register):
    """Represents a string register
    Between S000 - S999."""

    def __init__(self, is_16bit, initial_address="S0000"):
        Register.__init__(self, is_16bit, initial_address)



"""Normal register module"""
from keppy.register import Register

class RegularRegister(Register):
    """Represents a normal register
    Between K0000 - K9999.
    Each address is a byte on an 8-bit device and a word on a 16-bit
    device.
    To get a bit, you specify KXXXX.Y on an 8-bit device
    and KXXXX.YY on a 16-bit device.
    Can be data type Word, Short, BCD,
    DWord, Long, LBCD, Float, LLong, QWord,
    Double, Date, Boolean"""

    def __init__(self, is_16bit, initial_address="K0000"):
        Register.__init__(self, is_16bit, initial_address)



"""Kepware tag group module"""
import json
from collections import OrderedDict

from keppy.tag import Tag
class TagGroup(object):
    """Represents a Kepware tag group"""
    def __init__(self, tag_group_dict):
        self._tag_group_dict = tag_group_dict
        self._tags = self.parse_tags()
        self._is_ignored = False
        self._sub_groups = []
        # process sub groups
        if "tag_groups" in self._tag_group_dict:
            for sub_group in self._tag_group_dict["tag_groups"]:
                self._sub_groups.append(TagGroup(sub_group))

    def parse_tags(self):
        """Parses tags in tag group"""
        tags = []
        try:
            for tag in self._tag_group_dict["tags"]:
                tags.append(Tag(tag))
        except:
            return tags
        return tags

    @property
    def tags(self):
        """Gets the tags of the tag group"""
        return self._tags

    @property
    def name(self):
        """Gets the name of the tag group"""
        if self._is_ignored:
            return ''
        return self._tag_group_dict["common.ALLTYPES_NAME"]
    
    @property
    def sub_groups(self):
        return self._sub_groups

    def as_dict(self):
        """Returns dictionary representation of the tag group"""
        return self._tag_group_dict

    def as_json(self):
        """Returns the stringified JSON representation of the Kepware
        tag group"""
        return json.dumps(OrderedDict(self._tag_group_dict))

    def set_name(self, name):
        self._tag_group_dict["common.ALLTYPES_NAME"] = name

    def update(self):
        """Updates the dictionary of the tag group"""
        if self._is_ignored or "tags" not in self._tag_group_dict:
            return
        for i in range(len(self._tag_group_dict["tags"])):
            tag_dict = self._tag_group_dict["tags"][i]
            for tag in self._tags:
                if tag.name == tag_dict["common.ALLTYPES_NAME"]:
                    self._tag_group_dict["tags"][i] = tag.as_dict()
                    break

        for i in range(len(self._sub_groups)):
            sub_group = self._sub_groups[i]
            sub_group.update()
            self._tag_group_dict["tag_groups"][i] = sub_group.as_dict()



"""Channel module"""
from keppy.device import Device

class Channel(object):
    """Represents a Kepware channel"""
    def __init__(self, channel_dict, is_sixteen_bit, ignore_list):
        self._channel_dict = channel_dict
        self._ignore_list = ignore_list
        self.set_driver_simulated()
        self._is_sixteen_bit = is_sixteen_bit
        self._devices = self.parse_devices()

    def each_device(self, work):
        """Perform work on each device.
        Work is a function whcich takes a device as a parameter"""
        for device in self._devices:
            work(device)

    def parse_devices(self):
        """Creates an array of Device objects from the channel"""
        devices = []
        for device in self._channel_dict["devices"]:
            devices.append(Device(device, self._is_sixteen_bit, self._ignore_list))
        return devices

    def set_driver_simulated(self):
        """Sets the channel driver to simulator"""
        self._channel_dict["servermain.MULTIPLE_TYPES_DEVICE_DRIVER"] = "Simulator"

    @property
    def devices(self):
        """Gets the channel devices"""
        return self._devices

    @property
    def name(self):
        """Gets the name of the device"""
        return self._channel_dict["common.ALLTYPES_NAME"]

    def as_dict(self):
        """Returns dictionary representation of the channel"""
        return self._channel_dict

    def update(self):
        """Updates the dictionary of the channel"""
        for device in self.devices:
            device.update()
        for i in range(len(self._channel_dict["devices"])):
            device_dict = self._channel_dict["devices"][i]
            for device in self._devices:
                if device.name == device_dict["common.ALLTYPES_NAME"]:
                    self._channel_dict["devices"][i] = device.as_dict()



"""Tag type group module"""

from enum import Enum

class SiemensTcpIpTagType(Enum):
    """Enum for siemens TCP / IP data types"""
    STRING = 0
    BOOLEAN = 1
    BYTE = 3
    SHORT = 4
    WORD = 5
    DWORD = 7
    FLOAT = 8
    BOOLEAN_ARRAY = 21
    SHORT_ARRAY = 24
    WORD_ARRAY = 25
    DWORD_ARRAY = 27
    REAL_ARRAY = 28



"""Simulator device module"""

from keppy.regular_register import RegularRegister
from keppy.string_register import StringRegister
from keppy.tag_type_siemens import SiemensTcpIpTagType

class SimulatorDevice(object):
    """Represents a simulator device"""

    def __init__(self, is_16bit, normal_register_initial_address="K0000", string_register_initial_address="S000"):
        self._normal_register = RegularRegister(is_16bit, normal_register_initial_address)
        self._string_register = StringRegister(is_16bit, string_register_initial_address)
        self._is_sixteen_bit = is_16bit
        self._tag_type_processor = {
            SiemensTcpIpTagType.BOOLEAN: self.process_boolean,
            SiemensTcpIpTagType.BOOLEAN_ARRAY: self.process_boolean_array,
            SiemensTcpIpTagType.BYTE: self.process_byte,
            SiemensTcpIpTagType.DWORD: self.process_dword,
            SiemensTcpIpTagType.DWORD_ARRAY: self.process_dword_array,
            SiemensTcpIpTagType.FLOAT: self.process_float,
            SiemensTcpIpTagType.REAL_ARRAY: self.process_real_array,
            SiemensTcpIpTagType.SHORT: self.process_short,
            SiemensTcpIpTagType.SHORT_ARRAY: self.process_short_array,
            SiemensTcpIpTagType.STRING: self.process_string,
            SiemensTcpIpTagType.WORD: self.process_word,
            SiemensTcpIpTagType.WORD_ARRAY: self.process_word_array
        }

    @property
    def normal_register(self):
        """Gets the normal register"""
        return self._normal_register

    @property
    def string_register(self):
        """Gets the string register"""
        return self._string_register

    @property
    def is_sixteen_bit(self):
        """Returns True if device is 16 bit"""
        return self._is_sixteen_bit

    def process_tag(self, tag):
        """Processes tag and detects which function to use"""
        try:
            self._tag_type_processor[tag.data_type](tag)
        except KeyError as ex:
            raise Exception('Tag type {0} not recognized for tag {1}'
                            .format(
                                tag.data_type,
                                tag.name),
                            ex)


    def process_boolean(self, tag):
        """Process Boolean type tags"""
        tag.set_address(self.normal_register.current_bit_address)
        self.normal_register.move_to_next_bit_address()

    def process_boolean_array(self, tag):
        """Process Boolean array type tags"""
        array_size = tag.get_array_size()
        tag.set_address(self.normal_register.get_array(array_size))
        if self.is_sixteen_bit:
            # each boolean address needs 1/16 byte
            self.normal_register.move_to_next_address((array_size / 16) + 1)
            return
        # each boolean address needs 1/8 byte
        self.normal_register.move_to_next_address((array_size / 8) + 1)

    def process_byte(self, tag):
        """Process byte type tags"""
        tag.set_address(self.normal_register.current_address)
        # each address needs 1 byte
        self.normal_register.move_to_next_address(1)

    def _process_32_bit_type(self, tag):
        tag.set_address(self.normal_register.current_address)
        if self.is_sixteen_bit:
            # each word address needs 4 bytes = 2 addresses
            self.normal_register.move_to_next_address(2)
            return
        # each word address needs 4 bytes = 4 addresses
        self.normal_register.move_to_next_address(4)

    def _process_16_bit_type(self, tag):
        tag.set_address(self.normal_register.current_address)
        if self.is_sixteen_bit:
            # each short address needs 2 bytes = 1 address
            self.normal_register.move_to_next_address(1)
            return
        # each short address needs 2 bytes = 2 addresses
        self.normal_register.move_to_next_address(2)

    def process_dword(self, tag):
        """Process dword type tags"""
        self._process_32_bit_type(tag)

    def process_float(self, tag):
        """Process float type tags"""
        self._process_32_bit_type(tag)

    def process_dword_array(self, tag):
        """Process dword array type tags"""
        array_size = tag.get_array_size()
        tag.set_address(self.normal_register.get_array(array_size))
        if self.is_sixteen_bit:
            # each double word address needs 4 bytes = 2 addresses
            self.normal_register.move_to_next_address(array_size * 2)
            return
        # each double word address needs 4 bytes = 4 addresses
        self.normal_register.move_to_next_address(array_size * 4)

    def process_real_array(self, tag):
        """Process real array type tags"""
        array_size = tag.get_array_size()
        tag.set_address(self.normal_register.get_array(array_size))
        if self.is_sixteen_bit:
            # each real address needs 2 bytes = 1 address
            self.normal_register.move_to_next_address(array_size)
            return
        # each real address needs 2 bytes = 2 addresses
        self.normal_register.move_to_next_address(array_size * 2)

    def process_short(self, tag):
        """Process short type tags"""
        self._process_16_bit_type(tag)

    def process_word(self, tag):
        """Process word type tags"""
        self._process_16_bit_type(tag)

    def _process_16_bit_array(self, tag):
        array_size = tag.get_array_size()
        tag.set_address(self.normal_register.get_array(array_size))
        if self.is_sixteen_bit:
            # each short address needs two bytes = 1 address
            self.normal_register.move_to_next_address(array_size)
            return
        # each short address needs two bytes = 2 addresses
        self.normal_register.move_to_next_address(array_size * 2)

    def process_short_array(self, tag):
        """Process short array type tags"""
        self._process_16_bit_array(tag)

    def process_word_array(self, tag):
        """Process short array type tags"""
        self._process_16_bit_array(tag)

    def process_string(self, tag):
        """Process string type tags"""
        tag.set_address(self.string_register.current_address)
        if self.is_sixteen_bit:
            # each string address needs 1 byte = 1/2 an address
            self.string_register.move_to_next_address(1)
            return
        # each string address needs 1 byte = 1 address
        self.string_register.move_to_next_address(1)



"""Kepware project module"""
import json
from collections import OrderedDict
from keppy.channel import Channel

class Project(object):
    """Represents a Kepware project"""
    def __init__(self, kepware_dict, is_sixteen_bit, ignore_list):
        self._ignore_list = ignore_list
        self._kepware_dict = kepware_dict
        self._project_dict = kepware_dict["Project"]
        self._is_sixteen_bit = is_sixteen_bit
        self._channels = self.parse_channels()

    @property
    def channels(self):
        """Gets the channels of the project"""
        return self._channels

    def parse_channels(self):
        """Creates an array of Channel objects from the project"""
        channels = []
        for channel in self._project_dict["channels"]:
            channels.append(Channel(channel, self._is_sixteen_bit, self._ignore_list))
        return channels

    def as_json(self):
        """Returns the stringified JSON representation of the Kepware
        project"""
        return json.dumps(OrderedDict(self._kepware_dict))

    def update(self):
        """Updates the dictionary of the project"""
        for channel in self.channels:
            channel.update()
        for i in range(len(self._project_dict["channels"])):
            channel_dict = self._project_dict["channels"][i]
            for channel in self.channels:
                if channel.name == channel_dict["common.ALLTYPES_NAME"]:
                    self._project_dict["channels"][i] = channel.as_dict()



"""Kepware tag module"""

import re
from keppy.tag_type_siemens import SiemensTcpIpTagType

class Tag(object):
    """Represents a kepware tag"""
    def __init__(self, tag_dict):
        self._tag_dict = tag_dict
        self.address = property(self.get_address, self.set_address)

    @property
    def name(self):
        """The name of the tag"""
        return self._tag_dict["common.ALLTYPES_NAME"]

    @property
    def data_type(self):
        """Gets tag data type"""
        return SiemensTcpIpTagType(int(self._tag_dict["servermain.TAG_DATA_TYPE"]))

    def get_address(self):
        """Gets tag address"""
        return self._tag_dict["servermain.TAG_ADDRESS"]

    def set_address(self, value):
        """Sets tag address"""
        self._tag_dict["servermain.TAG_ADDRESS"] = value

    def name_replace(self, to_replace, replacement):
        """Replaces part of tag name with new value"""
        self.name = self.name.replace(to_replace, replacement)

    def as_dict(self):
        """Returns dictionary representation of the tag"""
        return self._tag_dict

    def get_array_size(self):
        """Attempts to parse array size out of the address"""
        match = re.search(r"(?<=\[)\d+(?=\])", self.get_address())
        return int(match.group(0))

    def get_string_length(self):
        """Attempts to parse array size out of the address"""
        try:
            return self.get_array_size()
        except:
            match = re.search(r"(?<=\.)\d+", self.get_address())
            try:
                return int(match.group(0))
            except Exception as ex:
                raise Exception('Could not get string size of {0} address {1}'
                                .format(
                                    self.name,
                                    self.get_address()),
                                ex)



"""Setup KEP.py"""
from codecs import open
from os import path
from setuptools import setup, find_packages

HERE = path.dirname(path.realpath(__file__))
with open(path.join(HERE, 'README.md'), encoding='utf-8') as readme:
    LONG_DESCRIPTION = readme.read()

setup(
    name='keppy',
    version='0.0.4',
    install_requires=['Enum34'],
    description='A Python parser of Kepware projects',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/jmbeach/KEP.py',
    author='jmbeach',
    author_email='jaredbeachdesign@gmail.com',
    keywords='kepware parser',
    packages=find_packages('src'),
    package_dir={'': 'src'}
)



"""Base register module"""
def pad_zeroes(addr, n_zeroes):
    """Padds the address with zeroes"""
    if len(addr) < n_zeroes:
        return pad_zeroes("0" + addr, n_zeroes)
    return addr

def int_addr(addr):
    """Gets the integer representation of an address"""
    return int(addr[1:])

def next_addr(addr, i):
    """Gets address after the current + i"""
    str_addr = pad_zeroes(str(int_addr(addr) + i), len(addr[1:]))
    return addr[0] + str_addr

class Register(object):
    """Represents the register of a simulator"""

    def __init__(self, is_16bit, initial_address):
        self._is_16bit = is_16bit
        self._current_address = initial_address
        self._register_map = {}
        self._size_of_current_register_address = 4
        self.mark_address(initial_address, self._size_of_current_register_address)
        self._current_bit_address = ""
        self.move_to_next_bit_address()

    @property
    def current_address(self):
        """Gets the current constant address of the register."""
        return self._current_address

    @property
    def is_16bit(self):
        """Gets whether the device is 16-bit
        (or else 8-bit)"""
        return self._is_16bit

    @property
    def current_bit_address(self):
        """Gets the current bit address"""
        return self._current_bit_address

    def get_register_letter(self):
        """Gets the letter representing the register (R, K, or S)
        """
        return self._current_address[0]

    def mark_address(self, addr, size):
        """Marks address as being used in simulator"""
        i = 0
        while i < size:
            self._register_map[addr] = True
            i += 1

    def is_address_in_use(self, addr):
        """Returns value which determines if register address in use"""
        return self._register_map.get(addr)

    def next_address_avoid_collision(self, start_addr):
        """Finds the next address recursively which does not collide with any other address"""
        for i in range(1, self._size_of_current_register_address):
            str_addr = next_addr(start_addr, i)
            if self.is_address_in_use(str_addr):
                return self.next_address_avoid_collision(
                    next_addr(start_addr, self._size_of_current_register_address + 1))
        return next_addr(start_addr, self._size_of_current_register_address)

    def next_address(self):
        """Returns the next address after the current"""
        return self.next_address_avoid_collision(self._current_address)

    def move_to_next_address(self, size_of_current):
        """Moves the register's current address to the next available.
        size_of_current specifies how many bytes/words to skip"""
        self._size_of_current_register_address = size_of_current
        self._current_address = self.next_address()
        self.mark_address(self._current_address, size_of_current)

    def move_to_next_bit_address(self):
        """Moves to next available bit address position"""
        self._current_bit_address = self.next_bit_address()

    def get_array(self, array_size):
        """Gets an array address"""
        return "{0}[{1}]".format(self._current_address, array_size)

    def next_bit_address(self):
        """Gets the next boolean address"""
        if self._current_bit_address == "":
            if self._is_16bit:
                return "{0}.{1}".format(
                    self.next_address(),
                    "00")
            return "{0}.{1}".format(
                self.next_address(),
                "0")
        if self._is_16bit:
            bool_half = int(self._current_bit_address.split(".")[1])
            if bool_half < 4:
                register_half = self._current_bit_address.split(".")[0]
                return "{0}.{1}".format(
                    register_half,
                    pad_zeroes(str(bool_half + 1), 2))
            self.move_to_next_address(self._size_of_current_register_address)
            return "{0}.{1}".format(
                self.next_address(),
                "00")
        bool_half = int(self._current_bit_address.split(".")[1])
        if bool_half < 3:
            register_half = self._current_bit_address.split(".")[0]
            return "{0}.{1}".format(
                register_half,
                bool_half + 1)
        self.move_to_next_address(self._size_of_current_register_address)
        return "{0}.{1}".format(
            self.next_address(),
            "0")



"""KEP.py init"""
from keppy.channel import Channel
from keppy.constant_register import ConstantRegister
from keppy.device import Device
from keppy.project import Project
from keppy.register import Register
from keppy.regular_register import RegularRegister
from keppy.simulator_device import SimulatorDevice
from keppy.string_register import StringRegister
from keppy.tag import Tag
from keppy.tag_group import TagGroup
from keppy.tag_type_siemens import SiemensTcpIpTagType
